import requests
from bs4 import BeautifulSoup
from datetime import datetime, date
import json
import os
import re
from typing import Optional

# Configuration
JSON_FILE = "pgn_urls_by_day.json"
OUTPUT_DIR = "combined"
DATE_FORMAT = "%Y-%m-%d"

ENGINE_MAPPING = [
    ("0.14.0", "2018-06-01", "0000"),
    ("0.14.1", "2018-07-08", "0000"),
    ("0.16.0", "2018-07-20", "0000"),
    ("0.17.0", "2018-08-27", "0000"),
    ("0.18.0", "2018-09-30", "0000"),
    ("0.18.1", "2018-10-02", "0000"),
    ("0.19.0", "2018-11-19", "0000"),
    ("0.19.1.1", "2018-12-10", "0000"),
    ("0.19.1", "2018-12-10", "0000"),
    ("0.20.0", "2019-01-01", "0000"),
    ("0.20.1", "2019-01-07", "0000"),
    ("0.20.2", "2019-02-01", "0000"),
    ("0.21.0", "2019-03-08", "0000"),
    ("0.21.1", "2019-03-23", "3475"),
    ("0.21.2", "2019-06-09", "0000"),
    ("0.21.3", "2019-07-21", "3479"),
    ("0.21.4", "2019-07-28", "0000"),
    ("0.22.0", "2019-08-05", "3536"),
    ("0.23.0", "2019-12-01", "0000"),
    ("0.23.1", "2019-12-03", "3650"),
    ("0.23.2", "2019-12-31", "0000"),
    ("0.23.3", "2020-02-18", "0000"),
    ("0.24.0", "2020-03-11", "0000"),
    ("0.24.1", "2020-03-15", "0000"),
    ("0.25.0", "2020-04-28", "0000"),
    ("0.25.1", "2020-04-30", "3681"),
    ("0.26.0", "2020-07-02", "0000"),
    ("0.26.1", "2020-07-15", "0000"),
    ("0.26.2", "2020-09-02", "0000"),
    ("0.26.3", "2020-10-10", "3656"),
    ("0.27.0", "2021-02-21", "3490"),
    ("0.28.0", "2021-08-25", "3594"),
    ("0.28.1", "2021-12-08", "0000"),
    ("0.28.2", "2021-12-13", "3585"),
    ("0.29.0", "2022-12-13", "3695"),
    ("0.30.0", "2023-07-21", "3660"),
    ("0.31.0", "2024-06-16", "3698"),
    ("0.31.1", "2024-08-11", "3707"),
    ("0.31.2", "2024-10-20", "3691"),
]

ENGINE_MAPPING = [
    (version, datetime.strptime(eff_date, "%Y-%m-%d").date(), rating)
    for version, eff_date, rating in ENGINE_MAPPING
]
ENGINE_MAPPING.sort(key=lambda x: x[1])

def generate_pgn_urls_json():
    """Generates the JSON file with PGN URLs grouped by date."""
    url = "https://training.lczero.org/matches/?show_all=1"
    response = requests.get(url)
    soup = BeautifulSoup(response.content, 'html.parser')
    
    table = soup.find('table', class_='table table-striped table-sm')
    if not table:
        print("Error: Could not find matches table.")
        return
    
    rows = table.find('tbody').find_all('tr')
    pgn_urls = {}
    
    for row in rows:
        cols = row.find_all('td')
        if len(cols) < 10:
            continue
        
        link = cols[0].find('a')
        if not link:
            continue
        match_id = link['href'].split('/')[-1]
        
        run_num = cols[1].text.strip()
        
        timestamp = cols[-1].text.strip()
        try:
            dt = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S %z")
            date_str = dt.date().isoformat()
        except ValueError:
            continue
        
        pgn_url = f"https://storage.lczero.org/files/match_pgns/{run_num}/{match_id}.pgn"
        
        if date_str not in pgn_urls:
            pgn_urls[date_str] = []
        pgn_urls[date_str].append(pgn_url)
    
    with open(JSON_FILE, 'w') as f:
        json.dump(pgn_urls, f, indent=2)
    print(f"PGN URLs JSON updated at {JSON_FILE}")

def get_engine_info(target_date: date) -> tuple[str, str]:
    """Get engine version and rating for a given date."""
    selected = ENGINE_MAPPING[0]
    for entry in ENGINE_MAPPING:
        if entry[1] <= target_date:
            selected = entry
        else:
            break
    return selected[0], selected[2]

def validate_date(date_str: str) -> Optional[date]:
    """Validate and return date object if valid."""
    try:
        input_date = datetime.strptime(date_str, DATE_FORMAT).date()
        if input_date < datetime(2018, 6, 17).date():
            print("Date cannot be earlier than 2018-06-17")
            return None
        today = date.today()
        if input_date >= today:
            print("Date cannot be today or in the future")
            return None
        return input_date
    except ValueError:
        print("Invalid date format. Please use YYYY-MM-DD.")
        return None

def download_pgn(url: str) -> str:
    """Download PGN content from URL with error handling."""
    try:
        response = requests.get(url, timeout=15)
        response.raise_for_status()
        return response.text
    except requests.RequestException as e:
        print(f"Error downloading {url}: {str(e)}")
        return ""

def process_game(game_text: str, header_date: str, engine_version: str, engine_rating: str) -> str:
    """Process individual game with header transformations."""
    headers = []
    moves = []
    has_date = False
    
    in_headers = True
    for line in game_text.splitlines():
        if in_headers and line.startswith("["):
            if line.startswith("[Event"):
                headers.append('[Event "Lc0 Matches"]')
            elif line.startswith("[Site"):
                headers.append('[Site "storage.lczero.org INT"]')
            elif line.startswith("[Date"):
                headers.append(f'[Date "{header_date}"]')
                has_date = True
            elif line.startswith("[Round"):
                if re.match(r'^\[Round\s+"-"\]', line):
                    headers.append('[Round "?"]')
                else:
                    headers.append(line)
            elif line.startswith("[White"):
                headers.append(f'[White "Lc0 {engine_version}"]')
            elif line.startswith("[Black"):
                headers.append(f'[Black "Lc0 {engine_version}"]')
            elif line.startswith(("[WhiteElo", "[BlackElo")):
                continue
            else:
                headers.append(line)
        else:
            in_headers = False
            moves.append(line)
    
    if not has_date:
        headers.insert(0, f'[Date "{header_date}"]')
    
    if engine_rating != "0000":
        headers.extend([
            f'[WhiteElo "{engine_rating}"]',
            f'[BlackElo "{engine_rating}"]'
        ])
    
    return "\n".join(headers) + "\n" + "\n".join(moves)

def process_pgns(pgn_content: str, header_date: str, engine_version: str, engine_rating: str) -> str:
    """Process full PGN content with multiple games."""
    games = []
    current_game = []
    
    for line in pgn_content.splitlines():
        if line.startswith("[Event ") and current_game:
            games.append("\n".join(current_game))
            current_game = []
        current_game.append(line)
    
    if current_game:
        games.append("\n".join(current_game))
    
    processed_games = []
    for game in games:
        if "[Event " in game:
            processed = process_game(game, header_date, engine_version, engine_rating)
            processed_games.append(processed)
    
    return "\n\n".join(processed_games)

def main():
    """Main execution flow."""
    date_str = input("Enter date in YYYY-MM-DD format: ").strip()
    target_date = validate_date(date_str)
    if not target_date:
        return
    
    # Check existing JSON or generate new
    pgn_data = {}
    json_exists = os.path.exists(JSON_FILE)
    if json_exists:
        try:
            with open(JSON_FILE, "r") as f:
                pgn_data = json.load(f)
        except (json.JSONDecodeError, FileNotFoundError):
            print(f"Error reading {JSON_FILE}, regenerating...")
            json_exists = False
    
    if date_str not in pgn_data:
        print(f"Date {date_str} not found in JSON, generating/updating JSON...")
        generate_pgn_urls_json()
        try:
            with open(JSON_FILE, "r") as f:
                pgn_data = json.load(f)
        except FileNotFoundError:
            print("Failed to generate JSON file.")
            return
    
    urls = pgn_data.get(date_str)
    if not urls:
        print(f"No PGNs found for {date_str}")
        return
    
    engine_version, engine_rating = get_engine_info(target_date)
    header_date = target_date.strftime("%Y.%m.%d")
    file_suffix = date_str[2:].replace("-", "")
    
    os.makedirs(OUTPUT_DIR, exist_ok=True)
    output_path = os.path.join(OUTPUT_DIR, f"lc0-matches-{file_suffix}-raw.pgn")
    combined = []
    
    for url in urls:
        print(f"Downloading {url}")
        pgn_content = download_pgn(url)
        if pgn_content:
            processed = process_pgns(pgn_content, header_date, engine_version, engine_rating)
            combined.append(processed)
    
    if combined:
        with open(output_path, "w", encoding="utf-8") as f:
            f.write("\n\n".join(combined))
        print(f"\nCombined PGN saved to {output_path}")
    else:
        print("No valid PGN content downloaded")

if __name__ == "__main__":
    main()